/* 
 * Based in f2a by Ulrich Hecht <uli@emulinks.de>
 * Multiboot by Eli Curtz <eli@nuprometheus.com>
 * if2a patch by D. Gauchard <deyv@free.fr>
 * F2A Ultra support by Vincent Rubiolo <vincent.rubiolo@free.fr>
 * Licensed under the terms of the GNU Public License version 2
 */

#include "libf2a.h"
#include "f2aio.h"
#include "f2autils.h"
#include "f2ausb.h"
#include "f2acarts.h"

/* default values are needed in case autodection is not used,
 * 18(256KB) is the most compatible value
 */
#define DEFAULT_ROMBLOCKSIZE_BIT	18
#define DEFAULT_WRITEBLOCKSIZE_BIT	18

int f2a_sim = 0;
static f2arecvmsg rm;
static f2asendmsg sm;

#if OSX || __ppc__ // little endian

static int f2a_usb_write_msg (f2asendmsg* original_sm)
{
	f2asendmsg sm;
	// this sm is a copy of *original_sm in which ppc little endian
	// values are converted into usb firmware's big endian:

	sm.command = swap32(original_sm->command);
	sm.size = swap32(original_sm->size);
	sm.magic = swap32(original_sm->magic);
	sm.subcommand = swap32(original_sm->subcommand);
	sm.address = swap32(original_sm->address);
	sm.sizekb = swap32(original_sm->sizekb);
	
	return f2a_usb_write((char*)&sm, sizeof(sm));
}

#else // big endian

inline int f2a_usb_write_msg(f2asendmsg* sm)
{
	return f2a_usb_write((char*)sm, sizeof(*sm));
}

#endif // !OSX


enum cart_type_e f2a_get_type (int* size_mbits, int* write_block_size_log2, int* rom_block_size_log2)
{
	int i;
	char buffer [LOADUNIT];
	int shift = 21;
	int txtsize = 20;
	char text[txtsize + 1];
	enum cart_type_e type = CART_TYPE_UNDEF;
	
	*write_block_size_log2 = DEFAULT_WRITEBLOCKSIZE_BIT;
	*rom_block_size_log2 = DEFAULT_ROMBLOCKSIZE_BIT;
	
	if (f2a_sim > 1)
		return type;
	
	*size_mbits = 0;

	// read background info from OAM : text boxes are drawn using tiled mode
	sm.command=CMD_READDATA;
	sm.subcommand=SUBCMD_READDATA;
	sm.size=LOADUNIT;
	sm.magic=MAGIC_NUMBER;
	sm.address=GBA_OAM;
	sm.sizekb=sm.size/1024;

	if (f2a_usb_write_msg(&sm) == -1)
		return type;

	if (f2a_usb_read(buffer, LOADUNIT) == -1)
		return type;

	// Translate from sprite memory to ASCII (hopefully this is easy)
	for (i = 0; i < txtsize; i++)
		text[i] = (unsigned char)buffer[(i + shift) * 8 + 4] / 2;
	// Remove trailing spaces
	for (i--; i > 0 && text[i] == ' '; i--);
	text[i + 1] = 0;
	
	if (strstr(text, "turbo"))
		type = CART_TYPE_TURBO;
	else if (strstr(text, "pro"))
		type = CART_TYPE_PRO;
	else if (strstr(text, "ultra"))
		type = CART_TYPE_ULTRA;

	if (strstr(text, "64M"))
		*size_mbits = 64;
	else if (strstr(text, "128M"))
		*size_mbits = 128;
	else if (strstr(text, "256M"))
		*size_mbits = 256;
	else if (strstr(text, "512M"))
		*size_mbits = 512;
	else if (strstr(text, "1G"))
		*size_mbits = 1024;
	
	// FIXME: confirm this !
#if 0
	if (strstr(text, "-B") || strstr(text, "-b")) // which one, both ?
		*write_block_size_log2 = 17; // 128Kbits
	else
		*write_block_size_log2 = 18; // 256Kbits
	// this is important and has been proven to work (-A, -X) with a f2apro-256-B only so far
	*write_block_size_log2 += 1;
	// please update if you have more info about this
#else
	*write_block_size_log2 = 18;
#endif
	
	*rom_block_size_log2 = 15; //15=32KB, has to be 17(128KB) for Turbo carts ?
	
	if (f2a_verbose)
	{
		print("\tFound '%s' as cartridge descriptor\n", text);
		print("\tAutodetected type is '%s', size is %iMbits and write block size is %iKbits\n",
			cart_type_str(type),
			*size_mbits,
			1 << (*write_block_size_log2 - 10));
	}

	return type;
}

int f2a_boot (binware_s* multiboot, binware_s* splash, int usb_timeout_in_seconds)
{
	unsigned char ack[64];
	
	if (f2a_sim > 1)
		return 0;

	memset(&sm, 0, sizeof(sm));

	// Boot the GBA
	sm.command=CMD_MULTIBOOT1;
	if (f2a_usb_write_msg(&sm) == -1)
		return -1;
	
	// Increase timeout as booting depends on user manipulation
	if (usb_timeout_in_seconds)
		f2a_usb_timeout = usb_timeout_in_seconds * 1000;

	sm.command=CMD_MULTIBOOT2;
	sm.size = multiboot->size;
	if (f2a_usb_write_msg(&sm) == -1)
	{
		f2a_usb_timeout = USB_TIMEOUT; // restore usb timeout
		return -1;
	}

	// Send the multiboot image
	if (f2a_usb_write(multiboot->data, multiboot->size) == -1)
	{
		f2a_usb_timeout = USB_TIMEOUT; // restore usb timeout
		return -1;
	}

	f2a_usb_timeout = USB_TIMEOUT; // restore usb timeout
	
	// Compulsory F2A reply
	// NOTE: this is not seen in win32 usb logs
	//       this comes from original code from Ulrich Hecht <uli@emulinks.de>
	if (f2a_usb_read((char*)ack, sizeof(ack)) == -1)
		return -1;

	if (f2a_verbose >= 2)
	{
		print("Post-boot:\n");
		print_array_dual(ack, 64);
	}
	
	// Send background picture
	sm.command=CMD_WRITEDATA;
	sm.subcommand=SUBCMD_WRITEDATA;
	sm.size=splash->size;
	sm.magic=MAGIC_NUMBER;
	sm.address=GBA_VRAM;
	sm.sizekb=sm.size/1024;

	// Command
	if (f2a_usb_write_msg(&sm) == -1)
		return -1;
	
	// Actual data
	if (f2a_usb_write(splash->data, splash->size) == -1)
		return -1;

	/* 
	 * Give time to GBA to copy background image to VRAM, we would otherwise
	 * experience info request failures whereas boot had been reported to be 
	 * have occurred fine.
	 */
	//sleep(0.5);

	return 0;
}


/* 
 * Get info from F2A linker
 * This is used to know whether the linker is ready to accept commands
 *
 * Returns:
 *  0 if linker is ready to accept commands
 *  1 if linker is not ready/not initialized
 *  -1 if there was something wrong with the USB commands
 */
int f2a_info (void)
{
//	extern int f2a_usb_timeout;

	if (f2a_sim > 1)
		return 0;

	memset(&sm, 0, sizeof(sm));
	memset(&rm, 0, sizeof(rm));

	sm.command=CMD_GETINF;

	/* 
	 * Info command takes very little time (63 + 64 bytes)
	 * Tweak timeout accordingly
	 */
	f2a_usb_timeout = 250; // 250ms
	
	if (f2a_usb_write_msg(&sm) == -1)
		return -1;

	if (f2a_usb_read((char*)&rm, sizeof(rm)) == -1)
		return -1;

	// Restore timeout to its default value
	f2a_usb_timeout = USB_TIMEOUT;
	
	if (f2a_verbose > 1)
	{
		print("F2A info:\n");
		print_array_dual((char*)&rm, sizeof(rm));
	}

	if (rm.data[0] == 0x4)
		return 0; // ready to accept commands
	else
		return 1; // not yet initialized/ready
}

int f2a_writemem(unsigned char* rom, int base, int offset, int size, int blocksize, int overall_displayed_size)
{
	int i;
	f2asendmsg sm;

	if (f2a_verbose)
		print("Burning... (base=0x%x offset=0x%x size=0x%x)\n", base, offset, size);

	// flash the ROM

	// initialize command buffer
	memset(&sm, 0, sizeof(sm));
	sm.command=CMD_WRITEDATA;
	// From sniffer under windows (pro): 6: GameSRAM a: GameROM
	// ULTRA : only 7 for write and 6 for read same as command field
	// XXXFIXME RECHECK THIS FOR PRO
	//sm.subcommand=(base >= GBA_SRAM)? SUBCMD_WRITESRAM: SUBCMD_WRITEROM;
	sm.subcommand = (base >= GBA_ROM && base < GBA_SRAM)? SUBCMD_WRITEROM: SUBCMD_WRITERAM;
	sm.magic=MAGIC_NUMBER;
	

#if 0
// Initialize command buffer
memset(&sm, 0, sizeof(sm));
sm.command = CMD_WRITEDATA;
sm.size = fileSize;
sm.magic = MAGIC_NUMBER;
sm.subcommand = SUBCMD_WRITESRAM;
sm.address = GBA_EWRAM;
sm.sizekb = fileSize / 1024;
#endif

	sm.size = size;
	sm.address = base + offset;
	sm.sizekb = size >> 10;

	if (!f2a_sim)
		if (f2a_usb_write_msg(&sm) == -1)
		{
			printerr("error sending command\n");
			return -1;
		}

	for(i = 0; i < size; i += blocksize)
	{
		if (!f2a_sim)
		{
			if (f2a_verbose > 2)
				print("Writing 0x%x bytes at base 0x%x offset 0x%x\n", blocksize, base, offset + i);
			if (f2a_usb_write(&rom[i], blocksize) == -1)
			{
 				printerr("error sending data\n");
				return -1;
			}
		}
//		if (f2a_verbose)
		{
			if (overall_displayed_size <= 0)
				overall_displayed_size = size;
			print("\rWrote chunk %i/%i (#%i/#%i) ",
			      (i + blocksize) >> 17,
			      size >> 17,
			      (i + blocksize + offset) >> 17,
			      (overall_displayed_size + offset) >> 17);
			printflush();
		}
	}
//	if (f2a_verbose)
		print("\r");
	
	if(f2a_verbose)
		print("\nDone\n");

	return 0;
}

int f2a_readmem(unsigned char* data, int address, int size)
{
	int i;

	assert((size & (LOADUNIT - 1)) == 0);

	memset(&sm, 0, sizeof(sm));
	sm.command=CMD_READDATA;
	sm.magic=MAGIC_NUMBER;
	sm.subcommand=SUBCMD_READDATA;
	sm.address=address;
	sm.size=size;
	sm.sizekb=size/1024;

	if (f2a_sim > 1)
		return -1;
		
	if (f2a_usb_write_msg(&sm) == -1)
		return -1;
	
	for (i = 0; i < sm.size; i += LOADUNIT)
	{
		if (f2a_usb_read(data, LOADUNIT) == -1)
			return -1;
		data += LOADUNIT;
	}
	return 0;
}

int f2a_readmemtofile(char* file, int address, int size, enum read_type_e read_type)
{
	int i;
	FILE* f;
	unsigned char data[LOADUNIT];

	assert((size & (LOADUNIT - 1)) == 0);

	memset(&sm, 0, sizeof(sm));

	if (f2a_sim > 1)
		return -1;
		
	switch (read_type)
	{
	case READ_MANY: // to read several memory areas into file
		if ((f = fopen(file, "ab")) == NULL)
		{
			printerrno(file);
			return -1;
		}
		break;
		
	case READ_ONCE:
		if ((f = fopen(file, "wb")) == NULL)
		{
			printerrno(file);
			return -1;
		}
		break;
		
	default:
		printerr("This is bad. Should have reached either WRITE_MANY or WRITE_ONCE\n");
		return -1;
	}

	sm.command=CMD_READDATA;
	sm.magic=MAGIC_NUMBER;
	sm.subcommand=SUBCMD_READDATA;
	sm.address=address;
	sm.size=size;
	sm.sizekb=size/1024;
	if (f2a_usb_write_msg(&sm) == -1)
		return -1;
	
//XXXFIXME use readfile(), write at once
	for (i = 0; i < sm.size; i += LOADUNIT)
	{
		if (f2a_usb_read(data, LOADUNIT) == -1)
		{
			fclose(f);
			return -1;
		}
		if (fwrite(data, LOADUNIT, 1, f) != 1)
		{
			printerrno(file);
			fclose(f);
			return -1;
		}
	}
	fclose(f);
	return 0;
}

int f2a_multiboot (char* fileName)
{
#if 0
	FILE*	file;
	size_t	fileSize;
	char	fileBuffer[LOADUNIT];
	struct stat st;
	int j;
   
	if (f2a_sim > 1)
		return 0;

	if (stat(fileName, &st) == -1)
	{
		printerrno(fileName);
		return -1;
	}
	fileSize = st.st_size;

	if ((file = fopen(fileName, "rb")) == NULL)
		printerrno(fileName);

	// Initialize command buffer
	memset(&sm, 0, sizeof(sm));
	sm.command = CMD_WRITEDATA;
	sm.size = fileSize;
	sm.magic = MAGIC_NUMBER;
	sm.subcommand = SUBCMD_WRITERAM;
	sm.address = GBA_EWRAM;
	sm.sizekb = fileSize / 1024;

	// Send the data to EWRAM
	if (f2a_usb_write_msg(&sm) == -1)
	{
		print("F2A Error: Unable to initiate Multiboot download\n");
		return -1;
	}

//XXXFIXME try and use readfile(), write at once
	for (j = 0; j < fileSize; j += LOADUNIT)
	{
		if (f2a_verbose > 0)
			print("Writing Multiboot data %d\n", (int)j);
	   
		// Read in LOADUNIT of data from the file
		if (fread(fileBuffer, LOADUNIT, 1, file) != 1)
		{
			printerrno(fileName);
			return -1;
		}
	   
		// Write LOADUNIT of data to the EWRAM
		if (f2a_usb_write(fileBuffer, LOADUNIT) == -1)
		{
			print("F2A Error: Unable to write Multiboot data\n");
			return -1;
		}
	}
	fclose(file);

#else
	// it is better like this :-) [remove this comment and the above #if 0]

	int	fileSize;
	char*	buffer;
   
	if (f2a_sim > 1)
		return 0;

	if ((buffer = download_from_file(fileName, &fileSize)) == NULL)
		return -1;

	if (f2a_writemem(buffer, GBA_EWRAM, 0, fileSize, fileSize, 0) < 0)
		return -1;
#endif   

	print("Now booting %s...\n", fileName);

	// Launch the program
	memset(&sm, 0, sizeof(sm));
	sm.command = CMD_WRITEDATA;
	sm.subcommand = SUBCMD_BOOTEWRAM;
	sm.magic = MAGIC_NUMBER;
	sm.address = GBA_EWRAM;
	   
	if (f2a_usb_write_msg(&sm) == -1)
	{
		print("F2A Error: Unable to initiate Multiboot download\n");
		return -1;
	}
   
	return 1;
} 

int f2a_burn (int cart_base, int cart_offset, char* rom, int rom_offset, int rom_size)
{
	// data addresses for f2a_burn() process must not cross MAXBURNCHUNK multiples
	int size_to_burn = rom_size;
	while (size_to_burn > 0)
	{
		// chunksize is the maximum size such that offset/MAXBURNCHUNK
		// and (offset+chunksize-1)/MAXBURNCHUNK are equals:
		int chunksize = (((cart_offset + rom_offset) + MAXBURNCHUNK) / MAXBURNCHUNK) * MAXBURNCHUNK - (cart_offset + rom_offset);
		if (chunksize > size_to_burn)
			chunksize = size_to_burn;

		int offset_burn = cart_offset + rom_offset;
		int size_burn = chunksize;
		int offset_save = offset_burn;
		int size_save = size_burn;

		if (f2a_verbose)
			print("Burning from 0x%x to 0x%x", cart_base + offset_burn, cart_base + offset_burn + size_burn);
		adjust_burn_addresses(&offset_burn, &size_burn);
		if ((offset_save != offset_burn) || (size_save != size_burn))
			print(" (adjusted: from 0x%x to 0x%x)\n", cart_base + offset_burn, cart_base + offset_burn + size_burn);
		else
			print("\n");

		if (f2a_writemem(&rom[rom_offset], cart_base, offset_burn, size_burn, f2a_write_block_size, rom_size) < 0)
			return -1;

		rom_offset += chunksize;
		size_to_burn -= chunksize;
	}
	return 0;
}

int old_f2a_burn (int base, char* rom, int offset, int size)
{
	// data addresses for f2a_burn() process must not cross MAXBURNCHUNK multiples
	while (size > 0)
	{
		// chunksize is the maximum size such that offset/MAXBURNCHUNK
		// and (offset+chunksize-1)/MAXBURNCHUNK are equals:
		int chunksize = ((offset + MAXBURNCHUNK) / MAXBURNCHUNK) * MAXBURNCHUNK - offset;
		if (chunksize > size)
			chunksize = size;

		int offset_burn = offset;
		int size_burn = chunksize;
		int offset_save = offset_burn;
		int size_save = size_burn;

		print("Burning from 0x%x to 0x%x", base + offset_burn, base + offset_burn + size_burn);
		adjust_burn_addresses(&offset_burn, &size_burn);
		if ((offset_save != offset_burn) || (size_save != size_burn))
			print(" (adjusted: from 0x%x to 0x%x)\n", base + offset_burn, base + offset_burn + size_burn);
		else
			print("\n");
		if (f2a_writemem(&rom[offset], base, offset_burn, size_burn, f2a_write_block_size, 0) < 0)
			return -1;

		offset += chunksize;
		size -= chunksize;
	}
	return 0;
}
