/* 
 * Based on f2a by Ulrich Hecht <uli@emulinks.de>
 * if2a patch by D. Gauchard <deyv@free.fr>
 * F2A Ultra support by Vincent Rubiolo <vincent.rubiolo@free.fr>
 * Firmware loading debug by Julien Janier <julien@janier.org>
 * Licensed under the terms of the GNU Public License version 2
 */

#include <usb.h>

#include "libf2a.h"
#include "f2ausb.h"
#include "f2aio.h"
#include "f2autils.h"
#include "an2131.h"

#if !_WIN32
#include <unistd.h>
#include <sys/wait.h>		// sleep
#endif

usb_dev_handle*		f2ahandle = NULL;
int			f2a_usb_timeout = USB_TIMEOUT;

//////////////////////////////////////////////////////////////////////////////
// linux-2.4.19- specifics
#if __linux__

#include <sys/utsname.h>	// uname()
#define EZUSB2131_MODULE "/proc/ezusb/dev0"

/* This function tests against the kernel version to know whether a USB reset 
 * and usb_set_configuration is needed (or unwanted).
 * If the machine runs linux-2.6, we do need to reset, if it is 2.4, we must not.
 * Also under 2.4, usb_set_configuration won't work whereas it is required
 * under win32.
 * 
 * We pass a verbosity parameter instead of relying on the global verbosity 
 * level because otherwise the kernel info would be printed each time we are 
 * called.
 */

static int f2a_linux24 (int verbosity)
{
	struct utsname utsname;
	int version, patchlevel;
	
	if (uname(&utsname) == -1)
	{
		printerrno("utsname");
		return 0;
	}

	version = atoi(strtok(utsname.release, "."));
	patchlevel = atoi(strtok(NULL, "."));

	if (verbosity == 1)
	{
		print("System informations:\n");
		print("\tSystem name = %s\n", utsname.sysname);
		print("\tNode name = %s\n", utsname.nodename);
		print("\tRelease = %s\n", utsname.release);
		print("\tVersion = %s\n", utsname.version);
		print("\tMachine = %s\n", utsname.machine);
	}

	if (strcmp(utsname.sysname, "Linux") == 0)
	{
		if (verbosity == 1)
			print("\tRunning on Linux %i.%i\n", version, patchlevel);
		return version == 2 && patchlevel == 4;
	}

	return 0;
}

static int ezusb_load_firmware_linux24 (const char* dev, const char* data, int size)
{
	int f;
	int w, tot;
	
	/* Trying to use ezusb2131 module (linux 2.4) */
	if ((f = open(dev, O_WRONLY)) == -1)
	{
		printerrno("open(%s)", dev);
		return -1;
	}
		
	if (f2a_verbose)
		print("Uploading EZ-USB firmware in %s...\n", dev);

	for (tot = 0; tot < size; tot += w)
	{
		w = write(f, data + tot, size - tot);
		if (w == -1)
		{
			printerrno("write on "EZUSB2131_MODULE);
			return -1;
		}
		
		if (f2a_verbose > 1)
			print(EZUSB2131_MODULE": wrote %i bytes [%i->%i / %i]\r", w, tot, tot+w, size);
	}

	close(f);

	if (f2a_verbose > 1)
		print("\n");

	return 0;
}

#else // !__linux__

#define f2a_linux24(x...) 0

#endif // !__linux__

//////////////////////////////////////////////////////////////////////////////
// static functions

static int f2a_usb_connect_root(void)
{
	struct usb_bus *bus;
	struct usb_device *dev;
	struct usb_device *f2a = NULL;

	int firmware_loaded=0;
	int init_hack = f2a_linux24(f2a_verbose > 1);

	usb_set_debug(f2a_verbose > 1);

	usb_init();
	usb_find_busses();

find_f2a:

	usb_find_devices();
	for (bus = usb_busses; !f2a && bus; bus = bus->next)
	{
		if (f2a_verbose)
			print("scanning bus '%s'\n", bus->dirname);
		for (dev = bus->devices; !f2a && dev; dev = dev->next)
		{
			if (f2a_verbose && (dev->descriptor.idVendor || dev->descriptor.idProduct))
				print("\tfound device '%s': 0x%04x/0x%04x\n", dev->filename, dev->descriptor.idVendor, dev->descriptor.idProduct);
			if ((dev->descriptor.idVendor == 0x547) && (dev->descriptor.idProduct == 0x1002))
			{
				f2a = dev;
				break;
			}
			else if ((dev->descriptor.idVendor == 0x547) && (dev->descriptor.idProduct == 0x2131) && !firmware_loaded)
			{
				printerr("Please turn OFF your GBA.\n");
				if (ezusb_load_firmware(dev, firmware.data) < 0
#if __linux__
				    && (init_hack && ezusb_load_firmware_linux24(EZUSB2131_MODULE, firmware.data, firmware.size) < 0)
#endif
				   )
				{
						return -1;
				}
			
				firmware_loaded = 1;
				if (f2a_verbose > 1)
					print("EZ-USB renumerating. Please wait\n");
				sleep(2);	// give the EZUSB some time to renumerate
				goto find_f2a;
			}
		}
	}
	if (f2a == NULL)
	{
		printerr("Error: couldn't find f2a attached to usb\n");
		return -1;
	}
	
	return !!(f2ahandle = f2a_usb_open(f2a, init_hack))? 0: -1;
}

//////////////////////////////////////////////////////////////////////////////
// public functions

usb_dev_handle* f2a_usb_open (struct usb_device* dev, int init_hack)
{
	usb_dev_handle* handle;
	
	// Open the USB device and get handle
	if ((handle = usb_open(dev)) == NULL)
	{
		printerr("usb_open: %s\n", usb_strerror());
		return NULL;
	}
	
	// Set configuration (needed for win32, unwanted in linux-2.4, no matter for linux-2.6)
	if (!init_hack && usb_set_configuration(handle, 1) < 0)
	{
		printerr("usb_set_configuration: %s\n", usb_strerror());
		usb_close(handle);
		return NULL;
	}

	// Reset if needed (yes for 2.6 kernels, no for 2.4 kernel, no matter for win32?)
	if (!init_hack && usb_reset(handle) < 0)
	{
		printerr("usb_reset: %s\n", usb_strerror());
		usb_close(handle);
		return NULL;
	}

	/* Be clean : claim interface. F2A linker has one interface (#0) and 
	 * 3 endpoints : 0x0 (control), 0x04 (to write into f2a), 0x83 (to 
	 * read)
	 */
	if (usb_claim_interface(handle, 0x0) < 0)
	{
		printerr("usb_claim_interface: %s\n", usb_strerror());
		usb_close(handle);
		return NULL;
	}

	return handle;
}


int f2a_usb_connect (void)
{
	int result = f2a_usb_connect_root();
#if !_WIN32
	seteuid(getuid());
#endif
	return result;
}

void f2a_usb_disconnect (void)
{
	/* F2A linker device has one interface (0x0)
	 * and 3 endpoints : 0x0 - control -, 0x4 - write -,
	 * 0x83 - read -
	 */
	if (usb_release_interface(f2ahandle, 0x0) < 0)
		printerr("usb_release: %s\n", usb_strerror());

	if (usb_close(f2ahandle) < 0)
		printerr("usb_close: %s\n", usb_strerror());
}

void f2a_usb_exit (int status)
{
	if (f2ahandle)
		f2a_usb_disconnect();
	exit(status);
}

int f2a_usb_read (char *buffer, int buffer_size)
{
	int result;

	result = usb_bulk_read(f2ahandle, 0x83, buffer, buffer_size, f2a_usb_timeout);
	if (result != buffer_size)
	{
		printerr("usb_bulk_read (request %i, got %i): %s\n", buffer_size, result, usb_strerror());
		return -1;
	}
	return result;
}

int f2a_usb_write(char *buffer, int buffer_size)
{
	int result;

	result = usb_bulk_write(f2ahandle, 0x4, buffer, buffer_size, f2a_usb_timeout);
	if (result != buffer_size)
	{
		printerr("usb_bulk_write (request %i, got %i): %s\n", buffer_size, result, usb_strerror());
		return -1;
	}
	return result;
}

int f2a_usb_write_by_block (char* buffer, int buffer_size, int block_size)
{
	int sent, result;
	
	sent = 0;
	while (sent < buffer_size)
	{
		result = f2a_usb_write(buffer + sent, buffer_size - sent > block_size? block_size: buffer_size - sent);
		if (result < 0)
			return -1;
		sent += result;
	}
	return sent;
}
