/* 
 * Based in f2a by Ulrich Hecht <uli@emulinks.de>
 * if2a patch by D. Gauchard <deyv@free.fr>
 * F2A Ultra support by Vincent Rubiolo <vincent.rubiolo@free.fr>
 * Licensed under the terms of the GNU Public License version 2
 */

// Main if2a file

#include <assert.h>

#include "getopt.h"
#include "libf2a.h"

void help (char* name)
{
	print(	"if2a " VERSION "\n"
		"Usage: %s options [file [file [...]]]\n"
		"Cart selection:\n"
		"	-t <t>  f2a cart type (T for Turbo, P for Pro, U for Ultra)\n"
		"	-S <s>	cart size (suffix kb,mb,kB,mB)\n"
		"	-L <f>	load loader or use an internal one (-L list)\n"
		"	-F <f>	load firmware or use an internal one (-F list)\n"
		"	-B <f>	load multiboot or use an internal one (-B list)\n"
		"	-I <f>	load splash image or an internal one (-I list)\n"
		"\nEASYROM;) options:\n"
		"	-A <r>	insert rom into cart (multiple options allowed)\n"
		"		<r>,<n> will change the name of the rom\n"
		"	-X <r>	remove rom from cart (match rom map - multiple allowed)\n"
		"	-p	show cart rom map\n"
		"\nROM options:\n"
		"	-R	read ROMs from cart (and generate filenames)\n"
		"		Individual ROM selection (optional):\n"
		"		.	: do not save ROM\n"
		"		@	: automatically generate filename\n"
		"		<name>	: assign <name> to ROM\n"
		"	-W	write ROMs to cart\n"
		"	-M	display ROM map\n"
		"	-s	do not try to reduce ROM size (default: reduce to fit loader)\n"
		"	-a	always try to reduce ROM size (default: reduce to fit loader)\n"
		"	-f	force writing (do not compare with cart contents)\n"
		"	-C	clean remaining ROMs\n"
		"\nSRAM options:\n"
		"	-r	read SRAM from cart\n"
		"	-w	write SRAM to cart\n"
		"\nDebugging options:\n"
		"	-c	(with -W) check whole file (not just borders) before burning\n"
		"	-n	do not insert f2a loader (*)\n"
		"	-H	do not check and correct ROM headers\n"
		"\nOther options:\n"
		"	-d	dummy - do not write -\n"
		"	-d	(one more -d) - do not read -\n"
		"	-v	be more f2a_verbose (Max verbosity is -vv)\n"
		"	-m <f>	send multiboot file\n"
		"\nLoader-PRO's (GBA-loader-3.x) SRAM manager specific:\n"
		"       -b <b>  specify block in SRAM\n"
		"	A block can be 'all', '1' or '2a' or '3b2' (same format as f2apro's cart loader)\n"
		"	1a1 (8KB) \\\n"
		"	1a2 (8KB)  \\  1a   \\\n"
		"	1a3 (8KB)  / (32KB) \\\n"
		"	1a4 (8KB) /          \\   1\n"
		"	1b1 (8KB) \\          / (64KB)\n"
		"	1b2 (8KB)  \\  1b    /\n"
		"	1b3 (8KB)  / (32KB)/\n"
		"	1b4 (8KB) /\n"
		"	2a1 (8KB)\n"
		"	...\n"
		"	4b4 (8KB)\n"
		"\nF2A ULTRA specific:\n"
		"	-u <f>	dump SVD (Saver Virtual Disk) data from cart into file\n"
		"	-U <f>	write SVD data to cart\n"
		"	-K <f>	write content descriptor to cart\n"
		"	-k <f>	dump content descriptor from cart into file\n"
		"	-G <f>	generate GameID for ROM\n"
		"	-D <f>	analyze content descriptor dump file\n"
		"	-E <f>	dump Die Hard data to file (early alpha)\n"
		"	-e <f>	write Die Hard data to cart (bogus function atm)\n"
		"\nNotes:\n"
		"+ Concerning rom files: you should always specify whole cart contents on cmdline,\n"
		"  cart will be updated according to differences between cmdline and cart contents\n"
		"+ *  Normally, f2a's loader is automatically inserted at beginning of cart\n",
		name);
}

int main (int argc, char** argv)
{
	int opt;
	enum mode_e mode = MODE_UNDEF;
	int play_with_map = 0;
	int enumerate_loaders = 0;
	char* block = NULL;
	char* multiboot_user_file = NULL;

	char* multiboot_file = NULL;
	char* firmware_file = NULL;
	char* splash_file = NULL;
	char* loader_file = NULL;
	
	char* add_files[FILE_NUMBER_MAX];
	int   add_files_number = 0;
	char* del_files[FILE_NUMBER_MAX];
	int   del_files_number = 0;
	
	int target_svd = 0;
	int target_dh = 0;
	int target_sram = 0;
	int target_rom = 0;
	int target_cd = 0;
	int map = 0;
	int clean_rom = 0;
	int gen_id = 0;
	int CD_dump = 0;
	enum cart_type_e cart_type = CART_TYPE_UNDEF;

	int sram_size = -1;
	int sram_offset = -1;

	f2a_init();

	if (argc == 1)
	{
		help(argv[0]);
		f2a_exit(5);
	}

	do
	{
		opt=getopt(argc, argv, "hMRrWwb:nfdvcsaS:m:HCt:eEuUkKGDL:F:B:I:pA:X:");
		switch(opt)
		{
		case 't':
			switch (optarg[0])
			{
			case 'P':
				cart_type = CART_TYPE_PRO;
				break;
			case 'T':
				cart_type = CART_TYPE_TURBO;
				break;
			case 'U':
				cart_type = CART_TYPE_ULTRA;
				break;
			default:
				printerr("cart type '%c' unknown\n", optarg[0]);
			}
			break;

		case 'E':
			target_dh = 1;
			mode = MODE_WRITE;
			break;

		case 'e':
			target_dh = 1;
			mode = MODE_READ;
			break;

		case 'U':
			target_svd = 1;
			mode = MODE_WRITE;
			break;

		case 'u':
			target_svd = 1;
			mode = MODE_READ;
			break;

		case 'K':
			target_cd = 1;
			mode = MODE_WRITE;
			break;

		case 'k':
			target_cd = 1;
			mode = MODE_READ;
			CD_dump = 1;
			break;
		
		case 'r':
			target_sram = 1;
			mode = MODE_READ;
			break;
			
		case 'R':
			target_rom = 1;
			mode = MODE_READ;
			break;
			
		case 'w':
			target_sram = 1;
			mode = MODE_WRITE;
			break;
			
		case 'W':
			target_rom = 1;
			mode = MODE_WRITE;
			break;
		
		case 'H':
			f2a_correct_header_allowed = 0;
			break;

		case 'C':
			clean_rom = 1;
			break;
		
		case 'b':
			block = optarg;
			break;

		case 'v':
			f2a_verbose++;
			break;
		
		case 'n':
			f2a_use_loader = 0;
			break;
		
		case 'f':
			f2a_force_burn = 1;
			break;
		
		case 'c':
			f2a_thorough_comp = 1;
			break;
		
		case 's':
			f2a_reduce_allowed = 0;
			break;
		
		case 'a':
			f2a_reduce_forced = 1;
			break;
		
		case 'S':
			f2a_cartsize_mb = convsize(optarg) >> 17;
			print("Setting cart size to %iMbits / %iKBytes\n",
				f2a_max_rom_size >> 17,
				f2a_max_rom_size >> 10);
			break;
			
		case 'd':
			f2a_sim++;
			break;
		
		case 'M':
			map = 1;
			break;
		
		case 'm':
			multiboot_user_file = optarg;
			break;

		case 'G':
			gen_id = 1;
			break;

		case 'D':
			target_cd = 1;
			mode = MODE_READ;
			CD_dump = 0;
			break;
		
		case 'L':
			loader_file = optarg;
			break;
		
		case 'F':
			firmware_file = optarg;
			break;
		
		case 'B':
			multiboot_file = optarg;
			break;
		
		case 'I':
			splash_file = optarg;
			break;
		
		case 'p':
			play_with_map = 1;
			break;
		
		case 'A':
			assert(add_files_number < FILE_NUMBER_MAX);
			add_files[add_files_number++] = optarg;
			play_with_map = 1;
			break;
		
		case 'X':
			assert(del_files_number < FILE_NUMBER_MAX);
			del_files[del_files_number++] = optarg;
			play_with_map = 1;
			break;

		case 'h':
		case '?':
			help(argv[0]);
			f2a_exit(1);
		}
	} while (opt!=-1);
	
	/* 
	 * Load default file if not already done by options.
	 * Loader is done later, once cart type is autodetected
	 */
	if (   select_firmware(firmware_file) < 0
	    || select_multiboot(multiboot_file) < 0
	    || select_splash(splash_file) < 0
	   )
	{
		printerr("Could not enumerate internal binaries (loader, firmware, etc)\n");
		f2a_exit(1);
	}
	
	enumerate_loaders = loader_file && strcmp(loader_file, "list") == 0;
	if (enumerate_loaders && cart_type != CART_TYPE_UNDEF)
	{
		select_loader(cart_type, loader_file);
		f2a_exit(1);
	}
	
	// Missing action
	if (mode == MODE_UNDEF && !map && !multiboot_user_file && !gen_id && !enumerate_loaders && !play_with_map)
	{
		printerr("No action specified\n");
		f2a_exit(1);
	}
	
	/* FIXME: check argument processing to ensure we really need this test 
	 * and could not replace some by getopt() argument check routines
	 * Instead of doing our own argument checking, we should use the ':' 
	 * flag for each option letter and use the OPTARG variable to get the 
	 * argument.
	 */
	if (argc-optind <= 0
	    && (   (mode == MODE_WRITE && target_rom)	// writing ROM needs filename
	        || target_sram				// may be unneeded 
	        || target_svd				// may be unneeded
	        || target_dh				// may be unneeded
	        || target_cd				// is unneeded
	        || gen_id))				// is unneeded
	{
		printerr("File name missing\n");
		f2a_exit(1);
	}

	// Prevent multiple arguments
	if ((target_sram || target_svd || target_dh || target_cd) && 
	    mode == MODE_READ && argc-optind > 1)
	{
		printerr("Cannot write SRAM, SVD, Die Hard or descriptor to multiple files\n");
		f2a_exit(1);
	}

	// Connection. This means parsing the USB bus and uploading firmware 
	if (f2a_connect() < 0)
	{
		printerr("Unable to connect to F2A linker\n");
		f2a_exit(1);
	}
	
	/* 
	 * Is the multiboot image already running?
	 * Ask the linker via the f2a_info() command to know.
	 */
	if (f2a_info() != 0)
	{
		// Error during USB command
#if 0
		if (f2a_info() == -1)
		{
			printerr("F2A linker had a problem replying to the info() command\n");
			f2a_exit(1);
		}
#endif
		// Boot
		printerr("Please turn OFF then ON your GBA with SELECT and START held down.\n");
		if (f2a_boot(&multiboot, &splash, 2000) < 0)
		{
			printerr("Cannot boot GBA\n");
			f2a_exit(1);
		}

		// Booting went well. We should be ready for commands
		if (f2a_info() != 0)
		{
			printerr("Booting went well but linker does not seem to be ready for commands. Exiting.\n");
			f2a_exit(1);
		}

		if (f2a_verbose)
			print("F2A multiboot image uploaded.\n");
	}

	// Cart type and parameters autodetection
	{
		int detect_f2a_cartsize_mb = 0;
		int detect_f2a_write_block_size_log2 = 0;
		enum cart_type_e detect_cart_type = f2a_get_type(&detect_f2a_cartsize_mb,
								 &detect_f2a_write_block_size_log2,
								 &f2a_rom_block_size_log2);
		
		if (cart_type == CART_TYPE_UNDEF)
			cart_type = detect_cart_type;
		else if (cart_type != detect_cart_type)
			printerr("Warning: Autodetected '%s' cart, but using '%s' as requested\n",
				 cart_type_str(detect_cart_type),
				 cart_type_str(cart_type));

		if (f2a_cartsize_mb <= 0)
			f2a_cartsize_mb = detect_f2a_cartsize_mb;
		else if (detect_f2a_cartsize_mb != f2a_cartsize_mb)
			printerr("Warning: Autodetected %iMbits cart, but using %iMbits as requested\n",
				 detect_f2a_cartsize_mb,
				 f2a_cartsize_mb);

		if (f2a_write_block_size_log2 <= 0)
			f2a_write_block_size_log2 = detect_f2a_write_block_size_log2;
		else if (detect_f2a_write_block_size_log2 != f2a_write_block_size_log2)
			printerr("Warning: Autodetected %iKbits write block size, but using %iKbits as requested\n",
				 1 << (detect_f2a_write_block_size_log2 - 10),
				 1 << (f2a_write_block_size_log2 - 10));
	}
	
	if (select_loader(cart_type, loader_file) < 0)
		f2a_exit(1);
	
	// SRAM block management
	if (cart_type == CART_TYPE_PRO && !block)
		block = ""; // same as "all" but not verbosed
	if (block && cart_type != CART_TYPE_ULTRA)
	{
		if (convf2ablock(block, &sram_offset, &sram_size) < 0)
		{
			printerr("Block %s has bad format\n", block);
			f2a_exit(1);
		}
	}
	else
	{
		if (cart_type != CART_TYPE_ULTRA && target_sram)
			print("No SRAM block specified, defaulting to 128KB SRAM\n");
		sram_offset = 0;
		sram_size = 131072 ; // = 128KBytes = 1Mbits
	}

	// Unsupported feature
	if ((target_cd || target_svd) && cart_type != CART_TYPE_ULTRA)
	{
		printerr("This feature is not supported by specified cart type\n");
		f2a_exit(1);
	}

	// Send multiboot
	if (mode == MODE_UNDEF && multiboot_user_file)
	{
		if (f2a_multiboot(multiboot_user_file) == -1)
		{
			printerr("Cannot send multiboot '%s' file to GBA\n", 
				 multiboot_user_file);
			f2a_exit(1);
		}
	}

	// Print cart contents
	if (mode == MODE_UNDEF && map)
	{
		if (old_display_cart_map() == -1)
			f2a_exit(1);
		if (load_cart_map() >= 0)
		{
			display_cart_map();
			reset_cart_map();
		}
	}
	
	// Write SRAM
	if (target_sram && mode == MODE_WRITE)
	{
		if (f2a_verbose)
			print("Uploading %iK to SRAM (offsets 0x%x -> 0x%x) from file %s\n", 
				(int) (sram_size/1024), 
				GBA_SRAM + sram_offset, 
				GBA_SRAM + sram_offset + sram_size - 1, 
				argv[optind]);
		
		assert(!target_cd);
		if (f2apro_loadandwrite_sram(argv[optind], sram_offset, sram_size) == -1)
		{
			printerr("Cannot read GBA SRAM\n");
			f2a_exit(1);
		}
	}
	
	// Read SRAM
	if (target_sram && mode == MODE_READ)
	{
		if (f2a_verbose)
			print("Downloading %iK from SRAM (offsets 0x%x -> 0x%x) to file %s\n",
				(int) (sram_size/1024), 
				GBA_SRAM + sram_offset, 
				GBA_SRAM + sram_offset + sram_size - 1, 
				argv[optind]);
				
		if (f2a_readmemtofile(argv[optind], GBA_SRAM + sram_offset, sram_size, 
					READ_ONCE) == -1)
		{
			printerr("Cannot read GBA SRAM\n");
			f2a_exit(1);
		}
	}
	
	// Write ROMs
	if (target_rom && mode == MODE_WRITE)
		auto_loadandburn_rom(cart_type, clean_rom, argc-optind, argv+optind);
	
	// Read ROMs
	if (target_rom && mode == MODE_READ)
		auto_readandsave_rom(argc-optind, argv+optind);

	// Write SVD
	if (target_svd && mode == MODE_WRITE)
	{
		if (f2a_verbose)
			print("Writing 320K of SVD data from file %s\n", 
				argv[optind]);
		f2au_SVD_from_file(argv[optind]);
	}
	
	// Read SVD
	if (target_svd && mode == MODE_READ)
	{
                if (f2a_verbose)
		        print("Reading 320K of SVD data to file %s\n", 
				argv[optind]);
		if (f2au_SVD_to_file(argv[optind]) == -1)
		{
			printerr("Cannot read GBA SVD\n");
			f2a_exit(1);
		}
	}

	// Write DH
	if (target_dh && mode == MODE_WRITE)
	{
		if (f2a_verbose)
			print("Writing XXK of Die Hard data from file %s\n", 
				argv[optind]);
		if (f2au_DH_from_file(argv[optind]) == -1)
		{
			printerr("Cannot write XXK or Die Hard data to GBA\n");
			f2a_exit(1);
		}
	}

	// Read DH
	if (target_dh && mode == MODE_READ)
	{
		if (f2a_verbose)
			print("Reading XXK of Die Hard data to file %s\n", 
				argv[optind]);
		if (f2au_DH_to_file(argv[optind]) == -1)
		{
			printerr("Cannot read GBA XXK or Die Hard data\n");
			f2a_exit(1);
		}
	}

	// Write CD
	if (target_cd && mode == MODE_WRITE)
	{
		if (f2a_verbose)
			print("Writing descriptor\n");
		if (f2au_loadandwrite_sram(argv[optind], F2AU_CD_BASE, SIZE_64K) == -1)
		{
			printerr("Cannot write GBA Descriptor\n");
			f2a_exit(1);
		}
	}
	
	// Read CD
	if (target_cd && mode == MODE_READ)
	{
		content_desc* desc;
		
		/* 
		 * Only thing that differs between dump and analysis is how to 
		 * get the buffer
		 */
		if (CD_dump)
		{
			if ((desc = f2au_CD_read()) == NULL)
				f2a_exit(1);
		}
		else
		{
			// FIXME : check for non-existent file
			buffer_from_file(argv[optind], (char*) &desc, 
					 sizeof(desc));
		}
		
		if (f2au_CD_check(desc) == -1)
		{
			printerr("Content descriptor seems invalid or corrupted. Exiting.\n");
			f2a_exit(1);
		}

		// CD dump request means no analysis
		if (CD_dump)
		{
			buffer_to_file(argv[optind], (char*) &desc, sizeof(desc));
			f2a_exit(0);
		}

		f2au_CD_print(desc);
	}
	
	// Generate GameID
        if (gen_id)
	{
		unsigned int game_id = 0;
		f2au_GameID_gen(argv[optind], &game_id);
		print("GameID for file %s is: %8X\n", argv[optind], game_id);
	}
	
	// show map, add or delete roms
	if (play_with_map)
	{
		if (load_cart_map() < 0)
			f2a_exit(1);
		
		if (del_files_number)
			cart_map_mark_for_remove(del_files, del_files_number);
		
		if (loader_file)
			cart_map_replace_loader(&loader);

		if ((del_files_number + add_files_number) == 0 || f2a_verbose > 0)
		{
			if (del_files_number)
				print("After removing files, ");
			display_cart_map();
		}
		
		if (del_files_number || add_files_number || loader_file)
		{
			cart_map_build_hole();
			if (f2a_verbose > 0)
				display_cart_map_hole();
		
			if (add_files_number)
			{
				if (cart_map_find_best_insertion_for_files(add_files, add_files_number) < 0)
				{
					reset_cart_map();
					f2a_exit(1);
				}
				if (f2a_verbose > 0)
					cart_map_file_display_best_score();
			}
		
			if (cart_map_process_changes() < 0)
			{
				reset_cart_map();
				f2a_exit(1);
			}
		}
			
		reset_cart_map();
	}
	
	// Cleanup after every action performed
	f2a_exit(0);

	return 0; // make compiler happy
}
